<?php
/**
 * @brief		Rules extension: Points
 * @package		Rules for IPS Social Suite
 * @since		11 Dec 2015
 * @version		SVN_VERSION_NUMBER
 */

namespace IPS\points\extensions\rules\Definitions;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * @brief	Rules definitions extension: Points
 */
class _Points
{

	/**
	 * @brief	The default option group title to list events, conditions, and actions from this class
	 */
	public $defaultGroup = 'Points Economy';

	/**
	 * Triggerable Events
	 *
	 * Define the events that can be triggered by your application
	 *
	 * Trigger events in your application code like this:
	 * if ( \IPS\Application::appIsEnabled( 'rules' ) )
	 * {
	 * 	\IPS\rules\Event::load( 'points', 'Points', 'event_key' )->trigger( $arg1, $arg2, etc... );
	 * }
	 *
	 * @return 	array		Array of event definitions
	 */
	public function events()
	{
		$events = array
		(
			'balance_adjusted' => array
			(
				'arguments' => array
				( 
					'balance' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Balance',
					),
					'amount' => array
					(
						'argtype' => 'float',
					),
				),
			),
			'transfer' => array
			(
				'arguments' => array
				(
					'source_balance' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Balance',
					),
					'destination_balance' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Balance',
					),
					'debit_amount' => array
					(
						'argtype' => 'float',
					),
					'credit_amount' => array
					(
						'argtype' => 'float',
					),
					'transfer' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Transaction\Transfer',
					),
				),
			),
			'credit_expired' => array
			(
				'arguments' => array
				(
					'balance' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Balance',
					),
					'amount' => array
					(
						'argtype' => 'float',
					),
				),
			),
			'interest_disbursed' => array
			( 
				'arguments' => array
				( 
					'balance' => array
					(
						'argtype' => 'object',
						'class' => 'IPS\points\Balance',
					),
					'amount' => array
					(
						'argtype' => 'float',
					),
				),
			),
		);
		
		return $events;
	}
	
	/**
	 * Conditional Operations
	 *
	 * You can define your own conditional operations which can be
	 * added to rules as conditions.
	 *
	 * @return 	array		Array of conditions definitions
	 */
	public function conditions()
	{
		$conditions = array
		(
			'balance_permitted' => array
			(
				'callback' 	=> array( $this, 'balancePermitted' ),
				'arguments'	=> array
				(
					'member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
				),				
			),
			'balance_check' => array
			(
				'callback' 	=> array( $this, 'balanceCheck' ),
				'configuration' => array
				(
					'form' => function( $form, $values )
					{
						$balance_types = array
						(
							'total' => 'points_balance_check_type_total',
							'available' => 'points_balance_check_type_available',
							'holding' => 'points_balance_check_type_holding',
							'unclaimed' => 'points_balance_check_type_unclaimed',
							'unpaid' => 'points_balance_check_type_unpaid',
						);
						
						$compare_methods = array
						(
							'<' => 'Less than',
							'>' => 'Greater than',
							'<=' => 'Less than or equal to',
							'>=' => 'Greater than or equal to',
							'==' => 'Equal to',
						);
					
						$form->add( new \IPS\Helpers\Form\Radio( 'points_balance_check_type', $values[ 'points_balance_check_type' ] ?: 'total', TRUE, array( 'options' => $balance_types ) ) );
						$form->add( new \IPS\Helpers\Form\Radio( 'points_balance_check_method', $values[ 'points_balance_check_method' ] ?: '<', TRUE, array( 'options' => $compare_methods ) ) );
					},
				),
				'arguments'	=> array
				(
					'member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
					'amount' => array
					(
						'argtypes' => array( 'int', 'float' ),
						'configuration' => array
						(
							'form' => function( $form, $values ) 
							{
								$form->add( new \IPS\Helpers\Form\Number( 'points_rules_amount', $values[ 'points_rules_amount' ], TRUE, array( 'min' => NULL, 'decimals' => 5 ), NULL, NULL, NULL, 'points_rules_amount' ) );
								return array( 'points_rules_amount' );
							},							
							'getArg' => function( $values, $action )
							{
								return $values[ 'points_rules_amount' ];
							},
						),
						'default' => 'manual',
						'required' => TRUE,
					),					
				),				
			),
			'transaction_permitted' => array
			(
				'callback' 	=> array( $this, 'transactionPermitted' ),
				'configuration' => array
				(
					'form' => function( $form, $values, $action )
					{
						$adjust_methods = array
						(
							'auto' => 'transaction_auto',
							'credit' => 'transaction_credit_only',
							'debit' => 'transaction_debit_only',
						);
						
						$form->add( new \IPS\Helpers\Form\Radio( 'points_rules_adjust_method', $values[ 'points_rules_adjust_method' ] ?: 'credit', TRUE, array( 'options' => $adjust_methods ) ) );
					},
				),
				'arguments'	=> array
				(
					'member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
					'amount' => array
					(
						'argtypes' => array( 'int', 'float' ),
						'configuration' => array
						(
							'form' => function( $form, $values ) 
							{
								$form->add( new \IPS\Helpers\Form\Number( 'points_rules_amount', $values[ 'points_rules_amount' ], TRUE, array( 'min' => NULL, 'decimals' => 5 ), NULL, NULL, NULL, 'points_rules_amount' ) );
								return array( 'points_rules_amount' );
							},							
							'getArg' => function( $values, $action )
							{
								return $values[ 'points_rules_amount' ];
							},
						),
						'default' => 'manual',
						'required' => TRUE,
					),
				),				
			),
			'transfer_permitted' => array
			(
				'callback' 	=> array( $this, 'transferPermitted' ),
				'arguments'	=> array
				(
					'source_member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'source_currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'source_bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
					'amount' => array
					(
						'argtypes' => array( 'int', 'float' ),
						'configuration' => array
						(
							'form' => function( $form, $values ) 
							{
								$form->add( new \IPS\Helpers\Form\Number( 'points_rules_amount', $values[ 'points_rules_amount' ], TRUE, array( 'min' => 0, 'decimals' => 5 ), NULL, NULL, NULL, 'points_rules_amount' ) );
								return array( 'points_rules_amount' );
							},							
							'getArg' => function( $values, $action )
							{
								return $values[ 'points_rules_amount' ];
							},
						),
						'default' => 'manual',
						'required' => TRUE,
					),
					'destination_member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member2' ),						
						'required'	=> TRUE,
					),
					'destination_currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency2' ),
						'required' => TRUE,
					),
					'destination_bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank2' ),
						'required' => FALSE,
					),
				),				
			),
			'check_member' => array
			(
				'callback' 	=> array( $this, 'checkMember' ),
				'configuration' => array
				(
					'form' => function( $form, $values )
					{
						$members = array();
						foreach( (array) $values[ 'points_rules_members' ] as $member_id )
						{
							if ( $member_id )
							{
								try
								{
									$members[] = \IPS\Member::load( $member_id );
								}
								catch ( \Exception $e ) {}
							}
						}
						
						$form->add( new \IPS\Helpers\Form\Member( 'points_rules_members', $members, TRUE, array( 'multiple' => NULL ), NULL, NULL, NULL, 'points_rules_members' ) );
					},
					'saveValues' => function( &$values )
					{	
						$members = array();
						foreach ( (array) $values[ 'points_rules_members' ] as $member )
						{
							$members[] = $member->member_id;
						}
						$values[ 'points_rules_members' ] = $members;
					},
				),
				'arguments'	=> array
				(
					'balance' => array
					(
						'argtypes' => static::argPreset( 'balance' ),
						'required' => TRUE,
					),
				),
			),
			'check_currency' => array
			(
				'callback' 	=> array( $this, 'checkCurrency' ),
				'configuration' => array
				(
					'form' => function( $form, $values )
					{
						$form->add( new \IPS\Helpers\Form\Node( 'points_rules_currency', $values[ 'points_rules_currency' ], TRUE, array( 'class' => 'IPS\points\Currency', 'multiple' => TRUE ), NULL, NULL, NULL, 'points_rules_currency' ) );
					},
					'saveValues' => function( &$values )
					{
						$save_values = array();
						
						foreach( (array) $values[ 'points_rules_currency' ] as $currency )
						{
							if ( $currency instanceof \IPS\points\Currency )
							{
								$save_values[] = $currency->id;
							}
						}
						
						$values[ 'points_rules_currency' ] = $save_values;
					},
				),
				'arguments'	=> array
				(
					'balance' => array
					(
						'argtypes' => static::argPreset( 'balance' ),
						'required' => TRUE,
					),
				),
			),
			'check_bank' => array
			(
				'callback' 	=> array( $this, 'checkBank' ),
				'configuration' => array
				(
					'form' => function( $form, $values )
					{
						$form->add( new \IPS\Helpers\Form\Node( 'points_rules_bank', $values[ 'points_rules_bank' ], TRUE, array( 'class' => 'IPS\points\Bank', 'multiple' => TRUE, 'zeroVal' => 'points_rules_anybank' ), NULL, NULL, NULL, 'points_rules_bank' ) );
					},
					'saveValues' => function( &$values )
					{
						$save_values = array();
						
						foreach( (array) $values[ 'points_rules_bank' ] as $bank )
						{
							if ( $bank instanceof \IPS\points\Bank )
							{
								$save_values[] = $bank->id;
							}
							else
							{
								$save_values[] = $bank;
							}
						}
						
						$values[ 'points_rules_bank' ] = $save_values;
					},
				),
				'arguments'	=> array
				(
					'balance' => array
					(
						'argtypes' => static::argPreset( 'balance' ),
						'required' => TRUE,
					),
				),				
			),
		);
		
		return $conditions;
	}

	/**
	 * Triggerable Actions
	 *
	 * @return 	array		Array of action definitions
	 */
	public function actions()
	{
		$actions = array
		(
			'credit_debit' => array
			(
				'callback' 	=> array( $this, 'creditDebitPoints' ),
				'configuration' => array
				(
					'form' => function( $form, $values, $action )
					{
						$adjust_methods = array
						(
							'auto' => 'transaction_auto',
							'credit' => 'transaction_credit_only',
							'debit' => 'transaction_debit_only',
						);
						
						$form->add( new \IPS\Helpers\Form\Radio( 'points_rules_adjust_method', $values[ 'points_rules_adjust_method' ] ?: 'auto', TRUE, array( 'options' => $adjust_methods ) ) );
						$form->add( new \IPS\Helpers\Form\YesNo( 'points_rules_enforce_perms', isset( $values[ 'points_rules_enforce_perms' ] ) ? $values[ 'points_rules_enforce_perms' ] : TRUE, FALSE ) );
						$form->add( new \IPS\Helpers\Form\YesNo( 'points_rules_expiration_select', $values[ 'points_rules_expiration_select' ], FALSE, array( 'togglesOn' => array( 'rules_Content_attribute_compare_minutes', 'rules_Content_attribute_compare_hours', 'rules_Content_attribute_compare_days', 'rules_Content_attribute_compare_months', 'rules_Content_attribute_compare_years' ) ), NULL, NULL, NULL, 'points_rules_expiration_select' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_minutes', $values[ 'rules_Content_attribute_compare_minutes' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_minutes' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_hours', $values[ 'rules_Content_attribute_compare_hours' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_hours' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_days', $values[ 'rules_Content_attribute_compare_days' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_days' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_months', $values[ 'rules_Content_attribute_compare_months' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_months' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_years', $values[ 'rules_Content_attribute_compare_years' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_years' ) );
					},
				),
				'arguments'	=> array
				(
					'member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
					'amount' => array
					(
						'argtypes' => array( 'int', 'float' ),
						'configuration' => array
						(
							'form' => function( $form, $values ) 
							{
								$form->add( new \IPS\Helpers\Form\Number( 'points_rules_amount', $values[ 'points_rules_amount' ], TRUE, array( 'min' => NULL, 'decimals' => 5 ), NULL, NULL, NULL, 'points_rules_amount' ) );
								return array( 'points_rules_amount' );
							},							
							'getArg' => function( $values, $action )
							{
								return $values[ 'points_rules_amount' ];
							},
						),
						'default' => 'manual',
						'required' => TRUE,
					),
					'message' => array
					(
						'argtypes' => array( 'string' ),
						'configuration' => array
						(
							'form' => function( $form, $values )
							{
								$form->add( new \IPS\Helpers\Form\Text( 'points_rules_message', $values[ 'points_rules_message' ], FALSE, array(), NULL, NULL, NULL, 'points_rules_message' ) );
								return array( 'points_rules_message' );
							},
							'getArg' => function( $values )
							{
								return $values[ 'points_rules_message' ];
							},
						),
						'default' => 'manual',
						'required' => FALSE,
					),
				),
			),
			'transfer_exchange' => array
			(
				'callback' 	=> array( $this, 'transferExchangePoints' ),
				'configuration' => array
				(
					'form' => function( $form, $values, $action )
					{
						$form->add( new \IPS\Helpers\Form\YesNo( 'points_rules_enforce_perms', isset( $values[ 'points_rules_enforce_perms' ] ) ? $values[ 'points_rules_enforce_perms' ] : TRUE, FALSE ) );
						$form->add( new \IPS\Helpers\Form\YesNo( 'points_rules_expiration_select', $values[ 'points_rules_expiration_select' ], FALSE, array( 'togglesOn' => array( 'rules_Content_attribute_compare_minutes', 'rules_Content_attribute_compare_hours', 'rules_Content_attribute_compare_days', 'rules_Content_attribute_compare_months', 'rules_Content_attribute_compare_years' ) ), NULL, NULL, NULL, 'points_rules_expiration_select' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_minutes', $values[ 'rules_Content_attribute_compare_minutes' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_minutes' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_hours', $values[ 'rules_Content_attribute_compare_hours' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_hours' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_days', $values[ 'rules_Content_attribute_compare_days' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_days' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_months', $values[ 'rules_Content_attribute_compare_months' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_months' ) );
						$form->add( new \IPS\Helpers\Form\Number( 'rules_Content_attribute_compare_years', $values[ 'rules_Content_attribute_compare_years' ] ?: 0, TRUE, array(), NULL, NULL, NULL, 'rules_Content_attribute_compare_years' ) );
					},
				),
				'arguments'	=> array
				(
					'source_member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member' ),						
						'required'	=> TRUE,
					),
					'source_currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency' ),
						'required' => TRUE,
					),
					'source_bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank' ),
						'required' => FALSE,
					),
					'amount' => array
					(
						'argtypes' => array( 'int', 'float' ),
						'configuration' => array
						(
							'form' => function( $form, $values ) 
							{
								$form->add( new \IPS\Helpers\Form\Number( 'points_rules_amount', $values[ 'points_rules_amount' ], TRUE, array( 'min' => 0.00001, 'decimals' => 5 ), NULL, NULL, NULL, 'points_rules_amount' ) );
								return array( 'points_rules_amount' );
							},							
							'getArg' => function( $values, $action )
							{
								return $values[ 'points_rules_amount' ];
							},
						),
						'default' => 'manual',
						'required' => TRUE,
					),
					'message' => array
					(
						'argtypes' => array( 'string' ),
						'configuration' => array
						(
							'form' => function( $form, $values )
							{
								$form->add( new \IPS\Helpers\Form\Text( 'points_rules_message', $values[ 'points_rules_message' ], FALSE, array(), NULL, NULL, NULL, 'points_rules_message' ) );
								return array( 'points_rules_message' );
							},
							'getArg' => function( $values )
							{
								return $values[ 'points_rules_message' ];
							},
						),
						'default' => 'manual',
						'required' => FALSE,
					),
					'destination_member' => array
					(
						'argtypes' => \IPS\rules\Application::argPreset( 'member' ),
						'configuration' => \IPS\rules\Application::configPreset( 'member', 'rules_choose_member2' ),						
						'required'	=> TRUE,
					),
					'destination_currency' => array
					(
						'argtypes' => static::argPreset( 'currency' ),
						'configuration' => static::configPreset( 'currency', 'points_rules_currency2' ),
						'required' => TRUE,
					),
					'destination_bank' => array
					(
						'argtypes' => static::argPreset( 'bank' ),
						'configuration' => static::configPreset( 'bank', 'points_rules_bank2' ),
						'required' => FALSE,
					),
				),
			),			
		);
		
		return $actions;
	}
	
	/**
	 * Arg Presets
	 *
	 * @param	string		$type		Type of preset to return
	 * @return	array
	 */
	public static function argPreset( $type )
	{
		switch( $type )
		{
			case 'balance':
			
				return array
				(
					'object' => array
					(
						'class' => 'IPS\points\Balance',
						'description' => 'A points balance', 
					),
				);
			
			case 'currency':

				return array
				(
					'object' => array
					(
						'class' => 'IPS\points\Currency',
						'description' => 'A points currency',
					),
				);
			
			case 'bank':
			
				return array
				(
					'object' => array
					(
						'class' => 'IPS\points\Bank',
						'description' => 'A points bank',
					),
				);
		}
	}
	
	/**
	 * Config Preset
	 *
	 * @param	string		$type		Type of preset to return
	 * @return	array
	 */
	public static function configPreset( $type, $field_name, $required=TRUE )
	{
		switch( $type )
		{
			case 'currency':

				return array
				(
					'form' => function( $form, $values ) use ( $field_name, $required )
					{
						$form->add( new \IPS\Helpers\Form\Node( $field_name, $values[ $field_name ], $required, array( 'class' => 'IPS\points\Currency' ), NULL, NULL, NULL, $field_name ) );
						return array( $field_name );
					},
					'saveValues' => function( &$values ) use ( $field_name )
					{
						if ( $values[ $field_name ] instanceof \IPS\points\Currency )
						{
							$values[ $field_name ] = $values[ $field_name ]->id;
						}
					},
					'getArg' => function( $values ) use ( $field_name )
					{
						return \IPS\points\Currency::load( $values[ $field_name ] );
					},
				);
			
			case 'bank':
			
				return array
				(
					'form' => function( $form, $values ) use ( $field_name, $required )
					{
						$form->add( new \IPS\Helpers\Form\Node( $field_name, $values[ $field_name ] ?: 0, $required, array( 'class' => 'IPS\points\Bank', 'zeroVal' => 'points_unbanked' ), NULL, NULL, NULL, $field_name ) );
						return array( $field_name );
					},
					'saveValues' => function( &$values ) use ( $field_name )
					{
						if ( $values[ $field_name ] instanceof \IPS\points\Bank )
						{
							$values[ $field_name ] = $values[ $field_name ]->id;
						}
					},
					'getArg' => function( $values ) use ( $field_name )
					{
						if ( $values[ $field_name ] == 0 )
						{
							return 0;
						}
						
						return \IPS\points\Bank::load( $values[ $field_name ] );
					},
				);
		}
	}
	
	/**
	 * Check if a balance is permitted
	 *
	 * @param	\IPS\Member|int			$member		Member to adjust balance for
	 * @param	\IPS\points\Currency|int	$currency	Currency to adjust balance for
	 * @param	\IPS\points\Bank|int		$bank		Bank to adjust balance for
	 * @return	bool
	 */
	public function balancePermitted( $member, $currency, $bank, $values )
	{
		try
		{
			$balance = \IPS\points\Balance::get( $member, $currency, $bank );
		}
		catch( \IPS\points\Exception $e )
		{
			return FALSE;
		}
		
		return TRUE;
	}
	
	/**
	 * Check the balance for a member
	 *
	 * @param	\IPS\Member|int			$member		Member to adjust balance for
	 * @param	\IPS\points\Currency|int	$currency	Currency to adjust balance for
	 * @param	\IPS\points\Bank|int		$bank		Bank to adjust balance for
	 * @return	bool
	 */
	public function balanceCheck( $member, $currency, $bank, $amount, $values )
	{
		try
		{
			$balance = \IPS\points\Balance::get( $member, $currency, $bank );
		}
		catch( \IPS\points\Exception $e )
		{
			return FALSE;
		}
		
		$balance_amount = 0;
		
		switch( $values[ 'points_balance_check_type' ] )
		{
			case 'total':
			
				$balance_amount = $balance->amount;
				break;
			
			case 'available':
			
				$balance_amount = $balance->available;
				break;
				
			case 'holding':
			
				$balance_amount = $balance->holding;
				break;
			
			case 'unclaimed':
			
				$balance_amount = $balance->unclaimed;
				break;
			
			case 'unpaid':
			
				$balance_amount = $balance->unpaid;
				break;
		}
		
		switch( $values[ 'points_balance_check_method' ] )
		{
			case '<': return $balance_amount < $amount;
			case '>': return $balance_amount > $amount;
			case '<=': return $balance_amount <= $amount;
			case '>=': return $balance_amount >= $amount;
			case '==': return $balance_amount == $amount;
		}
		
		return FALSE;
	}
	
	/**
	 * Check if a credit/debit can be made
	 *
	 * @param	\IPS\Member|int			$member		Member to adjust balance for
	 * @param	\IPS\points\Currency|int	$currency	Currency to adjust balance for
	 * @param	\IPS\points\Bank|int		$bank		Bank to adjust balance for
	 * @param	int|float			$amount		The amount for the transaction
	 * @return	bool
	 */
	public function transactionPermitted( $member, $currency, $bank, $amount, $values )
	{			
		try
		{
			$balance = \IPS\points\Balance::get( $member, $currency, $bank );
			
			if ( $amount == 0 )
			{
				return FALSE;
			}
			
			switch( $values[ 'points_rules_adjust_method' ] )
			{
				case 'auto':
				
					/* Detect adjustment method */
					if ( $amount >= 0 )
					{
						return $balance->canCredit( $amount );
					}
					else
					{
						return $balance->canDebit( $amount );
					}
				
				case 'credit': return $balance->canCredit( $amount );					
				case 'debit': return $balance->canDebit( $amount );
				default: return FALSE;
			}
		}
		catch( \IPS\points\Exception $e ) { }

		return FALSE;

	}
	
	/**
	 * Check if a transfer/exchange can be made
	 *
	 * @param	\IPS\Member|int			$source_member		Member being transferred from
	 * @param	\IPS\points\Currency|int	$source_currency	Currency being transferred from
	 * @param	\IPS\points\Bank|int		$source_bank		Bank being transferred from
	 * @param	int|float			$amount			The amount of the transfer
	 * @param	\IPS\Member|int			$destination_member	Member being transferred to
	 * @param	\IPS\points\Currency|int	$destination_currency	Currency being transferred to
	 * @param	\IPS\points\Bank|int		$destination_bank	Bank being transferred to
	 * @return	string
	 */
	public function transferPermitted( $source_member, $source_currency, $source_bank, $amount, $destination_member, $destination_currency, $destination_bank, $values )
	{
		try
		{
			$source = \IPS\points\Balance::get( $source_member, $source_currency, $source_bank );
			$destination = \IPS\points\Balance::get( $destination_member, $destination_currency, $destination_bank );
			
			return $source->canTransfer( $amount, $destination );
		}
		catch( \IPS\points\Exception $e )
		{
			return FALSE;
		}
		
		return TRUE;
	}
	
	/**
	 * Check a member
	 * 
	 * @return	bool
	 */
	public function checkMember( $balance, $values )
	{
		$members = (array) $values[ 'points_rules_members' ];
		
		if ( ! $balance instanceof \IPS\points\Balance )
		{
			return FALSE;
		}
		
		return in_array( $balance->member_id, $members );
	}
	
	/**
	 * Check a currency
	 *
	 * @return	bool
	 */
	public function checkCurrency( $balance, $values )
	{
		$currencies = (array) $values[ 'points_rules_currency' ];
		
		if ( ! $balance instanceof \IPS\points\Balance )
		{
			return FALSE;
		}
		
		return in_array( $balance->currency_id, $currencies );
	}
	
	/**
	 * Check a bank
	 *
	 * @return	bool
	 */
	public function checkBank( $balance, $values )
	{
		$banks = (array) $values[ 'points_rules_bank' ];
		
		if ( ( ! $balance instanceof \IPS\points\Balance ) or ( ! $balance->bank_id ) )
		{
			return FALSE;
		}
		
		return in_array( $balance->bank_id, $banks );
	}
	
	/**
	 * Credit/Debit Points
	 *
	 * @param	\IPS\Member|int			$member		Member to adjust balance for
	 * @param	\IPS\points\Currency|int	$currency	Currency to adjust balance for
	 * @param	\IPS\points\Bank|int		$bank		Bank to adjust balance for
	 * @param	int|float			$amount		The amount for the transaction
	 * @param	string				$message	The note for the transaction
	 * @return	string
	 */
	public function creditDebitPoints( $member, $currency, $bank, $amount, $message, $values )
	{
		try
		{
			$balance = \IPS\points\Balance::get( $member, $currency, $bank, $values[ 'points_rules_enforce_perms' ] );
			$expiration = 0;
			
			if ( $values[ 'points_rules_expiration_select' ] )
			{
				$value =  intval( $values[ 'rules_Content_attribute_compare_minutes' ] 	) * 60
					+ intval( $values[ 'rules_Content_attribute_compare_hours' ] 	) * ( 60 * 60 )
					+ intval( $values[ 'rules_Content_attribute_compare_days' ] 	) * ( 60 * 60 * 24 )
					+ intval( $values[ 'rules_Content_attribute_compare_months' ] 	) * ( 60 * 60 * 24 * 30 )
					+ intval( $values[ 'rules_Content_attribute_compare_years' ] 	) * ( 60 * 60 * 24 * 365 );
				
				if ( $value > 0 )
				{
					$expiration = time() + $value;
				}
			}
			
			if ( $amount == 0 )
			{
				return 'Transaction amount is zero';
			}
			
			switch( $values[ 'points_rules_adjust_method' ] )
			{
				case 'auto':
				
					/* Detect adjustment method */
					if ( $amount >= 0 )
					{
						$action = 'credited with';
						$balance->credit( $amount, $message, $expiration, array(), $values[ 'points_rules_enforce_perms' ] );
					}
					else
					{
						$action = 'debited for';
						$balance->debit( $amount, $message, $expiration, array(), $values[ 'points_rules_enforce_perms' ] );
					}
					break;
				
				case 'credit':
				
					if ( $amount < 0 )
					{
						return 'ignoring negative value';
					}
					
					$action = 'credited with';
					$balance->credit( $amount, $message, $expiration, array(), $values[ 'points_rules_enforce_perms' ] );
					break;
					
				case 'debit':
				
					if ( $amount > 0 )
					{
						return 'ignoring positive value';
					}
					
					$action = 'debited for';
					$balance->debit( $amount, $message, $expiration, array(), $values[ 'points_rules_enforce_perms' ] );
					break;
					
				default:
				
					return 'adjust method not configured for action';
			}
		}
		catch( \IPS\points\Exception $e )
		{
			return $e->getMessage();
		}

		return "balance id: {$balance->id} {$action} " . $balance->currency()->display( $amount ) . ' ' . $balance->currency()->pluralized( $amount ) . ( $expiration ? ' with an expiration of ' . \IPS\DateTime::ts( $expiration ) : '' );
	}
	
	/**
	 * Transfer/Exchange Points
	 *
	 * @param	\IPS\Member|int			$source_member		Member being transferred from
	 * @param	\IPS\points\Currency|int	$source_currency	Currency being transferred from
	 * @param	\IPS\points\Bank|int		$source_bank		Bank being transferred from
	 * @param	int|float			$amount			The amount of the transfer
	 * @param	string				$message		The note for the transaction
	 * @param	\IPS\Member|int			$destination_member	Member being transferred to
	 * @param	\IPS\points\Currency|int	$destination_currency	Currency being transferred to
	 * @param	\IPS\points\Bank|int		$destination_bank	Bank being transferred to
	 * @return	string
	 */
	public function transferExchangePoints( $source_member, $source_currency, $source_bank, $amount, $message, $destination_member, $destination_currency, $destination_bank, $values )
	{
		try
		{
			$source = \IPS\points\Balance::get( $source_member, $source_currency, $source_bank, $values[ 'points_rules_enforce_perms' ] );
			$destination = \IPS\points\Balance::get( $destination_member, $destination_currency, $destination_bank, $values[ 'points_rules_enforce_perms' ] );
			$expiration = 0;
			
			if ( $values[ 'points_rules_expiration_select' ] )
			{
				$value =  intval( $values[ 'rules_Content_attribute_compare_minutes' ] 	) * 60
					+ intval( $values[ 'rules_Content_attribute_compare_hours' ] 	) * ( 60 * 60 )
					+ intval( $values[ 'rules_Content_attribute_compare_days' ] 	) * ( 60 * 60 * 24 )
					+ intval( $values[ 'rules_Content_attribute_compare_months' ] 	) * ( 60 * 60 * 24 * 30 )
					+ intval( $values[ 'rules_Content_attribute_compare_years' ] 	) * ( 60 * 60 * 24 * 365 );
					
				if ( $value > 0 )
				{
					$expiration = time() + $value;
				}
			}
			
			$source->transfer( $amount, $destination, $message, $expiration, array(), $values[ 'points_rules_enforce_perms' ] );
		}
		catch( \IPS\points\Exception $e )
		{
			return $e->getMessage();
		}
		
		return "transfer/exchange of {$source->currency()->display( $amount )} {$source->currency()->pluralized( $amount )} completed between balances id: {$source->id} -> id: {$destination->id}" . ( $expiration ? ' with an expiration of ' . \IPS\DateTime::ts( $expiration ) : '' );
	}
	
	
}