<?php
/**
 * @brief		Member Points Systems
 * @author		Kevin Carwile (http://www.linkedin.com/in/kevincarwile)
 * @copyright		(c) 2015 - Kevin Carwile
 * @package		Points
 * @since		19 Nov 2015
 */

namespace IPS\points;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/* Make sure application class is loaded */
class_exists( 'IPS\points\Application' );

/**
 * Points Transaction
 */
class _Balance extends \IPS\Patterns\ActiveRecord
{

	/**
	 * @brief	[ActiveRecord] Multiton Store
	 */
	protected static $multitons;
	
	/**
	 * @brief	[ActiveRecord] Database Table
	 */
	public static $databaseTable = 'points_balances';
	
	/**
	 * @brief	[ActiveRecord] Database Prefix
	 */
	public static $databasePrefix = 'balance_';
	
	/**
	 * @brief	Points Balance Caches
	 */
	public static $balanceCache = array();
	
	/**
	 * Load Points Balance
	 *
	 * @param	\IPS\Member|int			$member		Member to get balance for
	 * @param	\IPS\points\Currency|int	$currency	Currency to get balance for
	 * @param	\IPS\points\Bank|int		$bank		Bank to get balance for
	 * @return	\IPS\points\Balance
	 *
	 * @throws	\IPS\points\Exception
	 */
	public static function get( $member, $currency, $bank=0, $enforcePerms=TRUE )
	{		
		/* Load Member */
		if ( ! $member instanceof \IPS\Member )
		{
			try
			{
				$member = \IPS\Member::load( $member );
			}
			catch( \OutOfRangeException $e )
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_bad_member' ) );
			}
		}
		
		/* Load Currency */
		if ( ! $currency instanceof \IPS\points\Currency )
		{
			try
			{
				$currency = \IPS\points\Currency::load( $currency );
			}
			catch( \OutOfRangeException $e )
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_bad_currency' ) );
			}
		}
		
		/* Load Bank */
		if ( ( ! $bank instanceof \IPS\points\Bank ) )
		{
			if ( is_numeric( $bank ) and $bank > 0 )
			{
				try
				{
					$bank = \IPS\points\Bank::load( $bank );
				}
				catch( \OutOfRangeException $e )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_bad_bank' ) );
				}
			}
		}
		
		/* Make sure member is not a guest */
		if ( ! $member->member_id )
		{
			throw new \OutOfRangeException( \IPS\Member::loggedIn()->language()->get( 'points_exception_guest_balance' ) );
		}
		
		if ( $enforcePerms )
		{
			/* Check that member has permission to have the balance */
			if ( $bank )
			{
				/* Check hard usage permission */
				if ( ! $bank->can( 'use', $member ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_bank_unavailable' ) );
				}
				
				/* Check currency is available in bank */
				if ( ! $bank->canUse( $currency, $member ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_bank_currency_unavailable' ) );
				}			
			}
			else
			{
				if ( ! $currency->can( 'use', $member ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_currency_unavailable' ) );
				}
			}
		}
		
		$bank_id = $bank ? $bank->id : 0;
		
		if ( isset( static::$balanceCache[ $member->member_id ][ $bank_id ][ $currency->id ] ) )
		{
			/* Grab cached balance */
			$balance = static::$balanceCache[ $member->member_id ][ $bank_id ][ $currency->id ];
		}
		else
		{
			/* Lookup the existing balance */
			try
			{
				$row = \IPS\Db::i()->select( '*', 'points_balances', array( 'balance_member_id=? AND balance_bank_id=? AND balance_currency_id=?', $member->member_id, $bank_id, $currency->id ) )->first();
				$balance = static::constructFromData( $row );
			}
			catch( \UnderflowException $e )
			{
				/* Create a new balance... */
				$balance = new \IPS\points\Balance;
				$balance->member_id = $member->member_id;
				$balance->bank_id = $bank_id;
				$balance->currency_id = $currency->id;
				$balance->amount = 0;
				$balance->last_accrual = time();
				$balance->last_cycle = time();
			}
		
			static::$balanceCache[ $member->member_id ][ $bank_id ][ $currency->id ] = $balance;
		}
		
		return $balance;
	}
	
	/**
	 * Check if this balance is available to member
	 *
	 * @return	bool
	 */
	public function isActive()
	{
		try
		{
			\IPS\points\Balance::get( $this->member_id, $this->currency_id, $this->bank_id );
			return TRUE;
		}
		catch( \Exception $e )
		{
			return FALSE;
		}
	}
	
	/**
	 * Get Member
	 *
	 * @return	\IPS\Member
	 */
	public function member()
	{
		return \IPS\Member::load( $this->member_id );
	}
	
	/**
	 * Get Currency
	 *
	 * @return	\IPS\points\Currency
	 */
	public function currency()
	{
		return \IPS\points\Currency::load( $this->currency_id );
	}
	
	/**
	 * Get Bank
	 *
	 * @return	\IPS\points\Bank|NULL
	 */
	public function bank()
	{
		if ( $this->bank_id )
		{
			return \IPS\points\Bank::load( $this->bank_id );
		}
		else
		{
			return NULL;
		}
	}
	
	/**
	 * Display An Amount
	 *
	 * @param	string		$register		The amount to display
	 * @return	string
	 */
	public function display( $register='amount' )
	{
		return $this->currency()->display( $this->$register );
	}
	
	/**
	 * Calculated available amount to spend/transfer
	 *
	 * @return	int|float
	 */
	public function get_available()
	{
		if ( ! $this->id or ! $this->bank() )
		{
			return $this->currency()->round( $this->amount );
		}
		
		$minimum = $this->bank()->bank_data[ 'bank_currency_minimum_' . $this->currency()->id ];
		$available = $this->currency()->round( $this->amount - $minimum - $this->holding );
		
		return $available > 0 ? $available : 0;
	}
	
	/**
	 * Calculate amount on hold
	 *
	 * @return	int|float
	 */
	public function get_holding()
	{
		if ( $this->id and $this->bank() )
		{		
			if ( $holding_period = $this->bank()->bank_data[ 'bank_currency_holding_' . $this->currency()->id ] )
			{
				$hold_since = (int) ( time() - ( $holding_period * ( 60 * 60 * 24 ) ) );
				$in_holding = \IPS\Db::i()->select( 'SUM(transaction_amount)', 'points_transactions', array( 'transaction_balance_id=? AND transaction_type=? AND transaction_timestamp>?', $this->id, \IPS\points\TRANSACTION_CREDIT, $hold_since ) )->first() ?: 0;
				
				return $this->currency()->round( $in_holding );
			}
		}

		return 0;
	}
	
	/**
	 * Calculate unclaimed credit total
	 *
	 * @return	int|float
	 */
	public function get_unclaimed()
	{
		$unclaimed_amount = \IPS\Db::i()->select( 'SUM(credit_amount)', 'points_credits', array( 'credit_member_id=? AND credit_currency_id=? AND credit_bank_id=? AND credit_expiration>? AND credit_transaction_id=0', $this->member_id, $this->currency_id, $this->bank_id, time() ) )->first();
		
		return $this->currency()->round( $unclaimed_amount );
	}
	
	/**
	 * Calculate unpaid debit total
	 *
	 * @return	int|float
	 */
	public function get_unpaid()
	{
		$unpaid_amount = \IPS\Db::i()->select( 'SUM(debit_amount)', 'points_debits', array( 'debit_member_id=? AND debit_currency_id=? AND debit_bank_id=? AND debit_expiration>0 AND debit_transaction_id=0', $this->member_id, $this->currency_id, $this->bank_id ) )->first();
		
		return $this->currency()->round( abs( $unpaid_amount ) );
	}
	
	/**
	 * @brief	Title
	 */
	public $_title = NULL;
	
	/**
	 * Get Title
	 *
	 * @return	string
	 */
	public function get_title()
	{
		if ( isset( $this->_title ) )
		{
			return $this->_title;
		}
		
		$title = ( $this->bank() ? $this->bank()->_title : \IPS\Member::loggedIn()->language()->addToStack( 'points_unbanked' ) ) . ' - ' . $this->currency()->_title;
		\IPS\Member::loggedIn()->language()->parseOutputForDisplay( $title );
		
		return $this->_title = $title;
	}
	
	/**
	 * Accrue interest & fees
	 *
	 * @return	void
	 */
	public function accrue()
	{
		/* Only banks support accrued points */
		if ( ! $this->bank() )
		{
			$this->last_accrual = time();
			$this->save();
			return;
		}
		
		/* Sanity Check */
		if ( $this->last_accrual == 0 )
		{
			$this->last_accrual = time();
		}
		
		/* Only accrue interest on balances if they are permitted */
		if ( $this->isActive() )
		{
			/* Accumulate accrued interest */
			if ( $this->bank()->bank_data[ 'bank_currency_interest_' . $this->currency()->id ] > 0 )
			{
				$interest_rate = $this->bank()->bank_data[ 'bank_currency_interest_' . $this->currency()->id ] / 100;
				$accrual_time = time() - $this->last_accrual;
				$interval_interest = $interest_rate * ( $accrual_time / ( 60 * 60 * 24 ) );
				$accrued_interest = $this->amount * $interval_interest;			
				$this->accrued_interest = round( $this->accrued_interest + $accrued_interest, 10 );
			}
			
			$cycle_interval = ( $this->bank()->cycle_interval ?: 1 ) * ( 60 * 60 * 24 );
			
			/* Disburse amounts on completed bank cycles */
			if ( time() >= $this->last_cycle + $cycle_interval )
			{
				/* Sanity check */
				if ( $this->last_cycle < ( time() - ( 60 * 60 * 24 * 1000 ) ) )
				{
					$this->last_cycle = time();
				}
				
				/* Advance cycle time by interval */
				while ( time() >= $this->last_cycle + $cycle_interval )
				{
					$this->last_cycle += $cycle_interval;
				}
				
				/* Disburse interest */
				$this->disburse();

				/* Charge bank fee */
				if ( $this->currency()->id == $this->bank()->fee_currency and $this->bank()->service_fee )
				{
					try
					{
						$this->debit( $this->bank()->service_fee, $this->member()->language()->get( 'points_bank_fee_charged' ), 0, array(), FALSE );
					}
					catch( \IPS\points\Exception $e ) { }
				}
				
			}
		}
		
		$this->last_accrual = time();
		$this->save();
	}
	
	/**
	 * Disburse accrued interest
	 *
	 * @return 	float		Amount disbursed
	 */
	public function disburse()
	{
		$disburse_amount = $this->currency()->smallest( $this->accrued_interest );
		
		if ( $disburse_amount != 0 )
		{
			try
			{
				if ( $disburse_amount > 0 )
				{
					$this->credit( $disburse_amount, $this->member()->language()->get( 'points_balance_interest_credited' ), 0, array(), FALSE );
				}
				else
				{
					$this->debit( $disburse_amount, $this->member()->language()->get( 'points_balance_interest_debited' ), 0, array(), FALSE );
				}
				
				$this->accrued_interest -= $disburse_amount;
				$this->save();
				
				/* Trigger rules event */
				if ( \IPS\Application::appIsEnabled( 'rules' ) )
				{
					\IPS\rules\Event::load( 'points', 'Points', 'interest_disbursed' )->trigger( $this, $disburse_amount );
				}
				
				return $disburse_amount;
			}
			catch( \IPS\points\Exception $e ) 
			{ 
				return 0;
			}
		}
	}
	
	/**
	 * Credit points
	 *
	 * @param	int|float			$amount			Amount of points to credit
	 * @param	string				$message		Description of transaction
	 * @param	int				$expiration		Optional expiration date of transaction if unclaimed
	 * @param	array				$data			Optional array of data to save with transaction
	 * @param	bool				$enforcePerms		Enforce permissions
	 * @return	\IPS\points\Transaction\Credit
	 *
	 * @throws	\IPS\points\Exception
	 */
	public function credit( $amount, $message='', $expiration=0, $data=array(), $enforcePerms=TRUE )
	{
		/* Round amount down to nearest currency precision */
		$amount = $this->currency()->smallest( $amount );
		
		/* Check for valid credit/debit amounts */
		if ( $amount == 0 )
		{
			throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_credit_zero' ) );
		}
		else if ( $amount < 0 )
		{
			throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_credit_negative' ) );
		}
		
		if ( $enforcePerms )
		{
			/* Check that credit can be made */
			$this->canCredit( $amount, TRUE );
		}
		
		/* Setup the credit */
		$credit = new \IPS\points\Transaction\Credit;
		$credit->amount = $amount;
		$credit->message = $message;
		$credit->data = json_encode( $data );
		$credit->timestamp = time();
		$credit->bank_id = $this->bank_id;
		$credit->currency_id = $this->currency_id;
		$credit->member_id = $this->member_id;
		
		if ( $expiration )
		{
			/* Save the credit to be claimed before expiration */
			$credit->expiration = $expiration;
			$credit->save();
		}
		else
		{
			/* Issue the credit immediately */
			$credit->save();
			$credit->issue( $message, array(), $enforcePerms );
		}
		
		return $credit;
	}
	
	/**
	 * Debit points
	 *
	 * @param	int|float			$amount			Amount of points to debit
	 * @param	string				$message		Description of transaction
	 * @param	int				$due_date		Optional date by which the amount can be paid
	 * @param	array				$data			Optional array of data to save with transaction
	 * @param	bool				$enforcePerms		Enforce permissions
	 * @return	\IPS\points\Transaction\Debit
	 *
	 * @throws	\IPS\points\Exception
	 */
	public function debit( $amount, $message='', $due_date=0, $data=array(), $enforcePerms=TRUE )
	{
		/* Round amount down to nearest currency precision */
		$amount = $this->currency()->smallest( $amount );
		
		/**
		 * Convert positive amounts to negative so this method can be used either way:
		 * > $balance->debit( 100 )
		 * > $balance->debit( -100 ) 
		 */
		if ( $amount > 0 )
		{
			$amount *= -1;
		}

		/* Check for valid credit/debit amounts */
		if ( $amount == 0 )
		{
			throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_debit_zero' ) );
		}

		if ( $enforcePerms )
		{
			/* Check that amount can be debited from balance */
			if ( ! $due_date )
			{
				$this->canDebit( $amount, TRUE );
			}
		}
		
		/* Setup the debit */
		$debit = new \IPS\points\Transaction\Debit;
		$debit->amount = $amount;
		$debit->message = $message;
		$debit->data = json_encode( $data );
		$debit->timestamp = time();
		$debit->bank_id = $this->bank_id;
		$debit->currency_id = $this->currency_id;
		$debit->member_id = $this->member_id;
		
		if ( $due_date > time() )
		{
			/* Save the debit to be paid before or on the due date */
			$debit->expiration = $due_date;
			$debit->save();
		}
		else
		{
			/* Issue the debit immediately */
			$debit->save();
			$debit->issue( $message, array(), $enforcePerms );
		}
		
		return $debit;
	}
	
	/**
	 * Transfer points
	 *
	 * @param	int|float		$amount			The amount to transfer from this balance
	 * @param	mixed			$destination		Where to send amount to, can be one of \IPS\points\Balance, \IPS\points\Currency, \IPS\points\Bank, \IPS\Member
	 * @param	string			$message		Optional message attached to transfer
	 * @param	int			$expiration		Optional expiration date of the transfer
	 * @param	array			$data			Data to attach to transfer
	 * @param	bool			$enforcePerms		Enforce permissions
	 * @return	void
	 *
	 * @throws	\IPS\points\Exception
	 */
	public function transfer( $amount, $destination, $message='', $expiration=0, $data=array(), $enforcePerms=TRUE )
	{
		if ( $destination instanceof \IPS\points\Balance )
		{
			/* No action required */
		}
		else if ( $destination instanceof \IPS\Member )
		{
			$destination = \IPS\points\Balance::get( $destination->member_id, $this->currency_id, $this->bank_id, $enforcePerms );
		}
		else if ( $destination instanceof \IPS\points\Currency )
		{
			$destination = \IPS\points\Balance::get( $this->member_id, $destination->id, $this->bank_id, $enforcePerms );
		}
		else if ( $destination instanceof \IPS\points\Bank )
		{
			$destination = \IPS\points\Balance::get( $this->member_id, $this->currency_id, $destination->id, $enforcePerms );
		}
		else
		{
			throw new \InvalidArgumentException( 'Invalid transfer destination' );
		}
		
		/* Check that transfer destination is not the same as the source */
		if 
		( 
			$this->bank_id == $destination->bank_id and
			$this->currency_id == $destination->currency_id and
			$this->member_id == $destination->member_id
		)
		{
			throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_invalid_transfer' ) );
		}
		
		/* Round amount down to nearest currency precision */
		$amount = $credit_amount = $this->currency()->smallest( $amount );		
		$transfer_type = \IPS\points\TRANSFER_NONE;
		
		/* Check if transfer is being made between banks */
		if ( $this->bank_id != $destination->bank_id )
		{
			$transfer_type = \IPS\points\TRANSFER_BANK;
		}
		
		/* Check if transfer requires a currency exchange */
		if ( $this->currency_id != $destination->currency_id )
		{
			$credit_amount = $this->currency()->exchange( $amount, $destination->currency() );
			$transfer_type = \IPS\points\TRANSFER_EXCHANGE;
		}
		
		/* Check if transfer is being sent between members */
		if ( $this->member_id != $destination->member_id )
		{
			$transfer_type = \IPS\points\TRANSFER_MEMBER;
		}
		
		/* Check for valid credit/debit amounts */
		if 		( $credit_amount == 0 ) 	{ throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_credit_zero' ) ); }
		else if		( $credit_amount < 0 ) 		{ throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_credit_negative' ) ); }
		else if		( $amount == 0 )		{ throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_debit_zero' ) );	}
		
		/* Check that the transfer doesn't violate any permissions */
		if ( $enforcePerms )
		{
			$this->canTransfer( $amount, $destination, TRUE, $expiration );
		}
		
		/* Execute the transfer */
		$debit = $this->debit( $amount, $message, 0, array(), FALSE );
		$credit = $destination->credit( $credit_amount, $message, $expiration, array(), FALSE );
		
		/* Log the transfer */
		$transfer = new \IPS\points\Transaction\Transfer;
		$transfer->debit_id = $debit->id;
		$transfer->credit_id = $credit->id;
		$transfer->type = $transfer_type;
		$transfer->data = json_encode( $data );
		$transfer->save();
		
		/* Link back the transfer credit/debit log */
		$debit->transfer_id = $transfer->id;
		$credit->transfer_id = $transfer->id;
		$debit->save();
		$credit->save();
		
		/* Trigger rules event */
		if ( \IPS\Application::appIsEnabled( 'rules' ) )
		{
			\IPS\rules\Event::load( 'points', 'Points', 'transfer' )->trigger( $this, $destination, $amount, $credit_amount, $transfer );
		}
		
	}
	
	/**
	 * Adjust balance +/-
	 *
	 * @param	int|float	$amount			The amount to adjust the balance
	 * @param	bool		$enforcePerms		Enforce permissions
	 * @return 	void
	 *
	 * @throws	\IPS\points\Exception
	 */
	public function adjust( $amount, $enforcePerms=TRUE )
	{
		if ( $enforcePerms )
		{
			/* Confirm that new balance is within limits */
			if ( $amount >= 0 )
			{
				$this->canCredit( $amount, TRUE );
			}
			else
			{
				$this->canDebit( $amount, TRUE );
			}
		}
		
		$this->amount += $amount;
		$this->save();
		
		/* Trigger rules event */
		if ( \IPS\Application::appIsEnabled( 'rules' ) )
		{
			\IPS\rules\Event::load( 'points', 'Points', 'balance_adjusted' )->trigger( $this, $amount );
		}
	}
	
	/**
	 * Can credit amount?
	 *
	 * @param	int|float	$amount		The amount to credit
	 * @param	bool		$throw		Throw an exception instead of returning a boolean value
	 * @return	bool
	 */
	public function canCredit( $amount, $throw=FALSE )
	{
		if ( $bank = $this->bank() )
		{
			$bank_data = $bank->bank_data;
			$maximum = isset( $bank_data[ 'bank_currency_maximum_' . $this->currency_id ] ) ? $bank_data[ 'bank_currency_maximum_' . $this->currency_id ] : -9999;
		}
		else
		{
			/* -9999 = no limit */
			$maximum = -9999;
		}

		if ( $maximum != -9999 and $this->amount + $amount > $maximum )
		{
			if ( $throw )
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_credit_maximum' ) );
			}
			
			return FALSE;
		}
		
		return TRUE;
	}
	
	/**
	 * Can debit amount?
	 *
	 * @param	int|float	$amount		The amount to debit
	 * @param	bool		$throw		Throw an exception instead of returning a boolean value
	 * @return 	bool
	 */
	public function canDebit( $amount, $throw=FALSE )
	{
		$amount = $amount < 0 ? $amount * -1 : $amount;
		$minimum = 0;
		
		if ( $bank = $this->bank() )
		{
			$bank_data = $bank->bank_data;
			$minimum = $bank_data[ 'bank_currency_minimum_' . $this->currency()->id ] ?: 0;
		}
		
		if ( $this->amount - $amount < $minimum )
		{
			if ( $throw )
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_debit_minimum' ) );		
			}
			
			return FALSE;
		}
		
		if ( $this->available < $amount )
		{
			if ( $throw )
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_balance_debit_insufficient' ) );
			}
			
			return FALSE;
		}

		return TRUE;
	}
	
	/**
	 * Can transfer/exchange amount?
	 *
	 * @param	int|float		$amount		The amount to debit (use 0 to check only if a transfer is possible)
	 * @param	\IPS\points\Balance	$destination	The balance to transfer/exchange to
	 * @param	bool			$throw		Throw an exception instead of returning a boolean value
	 * @param	int			$expiration	The expiration date for the transfer
	 * @return 	bool
	 */
	public function canTransfer( $amount, \IPS\points\Balance $destination, $throw=FALSE, $expiration=0 )
	{
		try
		{
			/* Round amount down to nearest currency precision */
			$amount = $credit_amount = $this->currency()->smallest( $amount );
			
			/* Check that transfer destination is not the same as the source */
			if 
			( 
				$this->bank_id == $destination->bank_id and
				$this->currency_id == $destination->currency_id and
				$this->member_id == $destination->member_id
			)
			{
				throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_invalid_transfer' ) );
			}
			
			/* Check if transfer is being made between banks */
			if ( $this->bank_id != $destination->bank_id )
			{
				if ( $this->bank_id and ! $this->bank()->can( 'transfer_out', $this->member() ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_transfer_out' ) );
				}
				
				if ( $destination->bank_id and ! $destination->bank()->can( 'transfer_in', $destination->member() ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_transfer_in' ) );
				}
			}
			
			/* Check if transfer requires a currency exchange */
			if ( $this->currency_id != $destination->currency_id )
			{
				if ( ! $this->currency()->can( 'exchange_out', $this->member() ) or ! $destination->currency()->can( 'exchange_in', $destination->member() ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_exchange_restriction' ) );
				}
				
				$credit_amount = $this->currency()->exchange( $amount, $destination->currency() );
			}
			
			/* Check if transfer is being sent between members */
			if ( $this->member_id != $destination->member_id )
			{
				if ( ! $this->currency()->can( 'send', $this->member() ) or ! $destination->currency()->can( 'receive', $destination->member() ) )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_send_receive' ) );
				}
			}
			
			/* By calling this method with amount=0, we can just check if a transfer is possible */
			if ( $amount !== 0 )
			{
				/* Check transfer doesn't exceed a credit limit */
				if ( ! $expiration )
				{
					$destination->canCredit( $credit_amount, TRUE );
				}
				
				/* Check transfer doesn't exceed a debit limit */
				$this->canDebit( $amount, TRUE );
			}
		}
		catch( \IPS\points\Exception $e )
		{
			if ( $throw )
			{
				throw $e;
			}
			
			return FALSE;
		}
		
		return TRUE;
	}
	
	/**
	 * Transactions table
	 *
	 * @param	int				$limit			Number of transactions to show per page
	 * @param	\IPS\Http\Url			$url			The url that the table is being displayed on
	 * @return	\IPS\Helpers\Table\Db
	 */
	public function transactionsTable( $limit=25, $url=NULL )
	{
		$url = $url ?: \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=transactions', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );

		/* Create the table */
		$table = new \IPS\Helpers\Table\Db( 'points_transactions', $url, array( array( 'transaction_balance_id=?', $this->id ) ) );
		
		$table->tableTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'table' );
		$table->rowsTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'rows' );
		
		$table->include = array( 'transaction_timestamp', 'transaction_type', 'transaction_message', 'transaction_amount', 'transaction_balance' );
		$table->langPrefix = 'table_';
		$table->sortBy = 'transaction_id';
		$table->sortDirection = 'DESC';
		$table->noSort = $table->include;
		$table->limit = $limit;
		$self = $this;
		
		$table->parsers = array
		(
			'transaction_timestamp' => function( $val, $row )
			{
				$datetime = \IPS\DateTime::ts( $val );
				return (string) $datetime;
			},
			'transaction_type' => function( $val, $row )
			{
				$lang = \IPS\Member::loggedIn()->language();
				$register = \IPS\points\Transaction\Register::load( $row[ 'transaction_id' ] );
				
				switch( $val )
				{
					case \IPS\points\TRANSACTION_CREDIT: 
					
						try
						{
							if ( $transfer = $register->transaction()->transfer() )
							{							
								switch( $transfer->type )
								{
									case \IPS\points\TRANSFER_BANK		: 
									
										try {
											$balance = $transfer->debit()->balance();
											$destination = "<a href='{$balance->url()}'>{$balance->bank()->name}</a>";
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_transfer_in', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
										
									case \IPS\points\TRANSFER_EXCHANGE	: 
									
										try {
											$balance = $transfer->debit()->balance();
											$destination = "<a href='{$balance->url()}'>{$balance->currency()->plural()}</a>";
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_exchange_in', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
										
									case \IPS\points\TRANSFER_MEMBER	: 
									
										try {
											$member = $transfer->debit()->member();
											$destination = $member->link();
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_member_in', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
								}
							}
						}
						catch( \OutOfRangeException $e ) { }
						
						return $lang->addToStack( 'transaction_credit' );
					
					case \IPS\points\TRANSACTION_DEBIT: 
					
						try
						{
							if ( $transfer = $register->transaction()->transfer() )
							{							
								switch( $transfer->type )
								{
									case \IPS\points\TRANSFER_BANK		: 
									
										try {
											$balance = $transfer->credit()->balance();
											$destination = "<a href='{$balance->url()}'>{$balance->bank()->name}</a>";
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_transfer_out', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
										
									case \IPS\points\TRANSFER_EXCHANGE	: 
									
										try {
											$balance = $transfer->credit()->balance();
											$destination = "<a href='{$balance->url()}'>{$balance->currency()->plural()}</a>";
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_exchange_out', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
										
									case \IPS\points\TRANSFER_MEMBER	: 
									
										try {
											$member = $transfer->credit()->member();
											$destination = $member->link();
										}
										catch( \Exception $e )
										{
											$destination = "{error}";
										}
										
										return $lang->addToStack( 'transaction_member_out', FALSE, array( 'htmlsprintf' => array( $destination ) ) );
								}
							}
						}
						catch( \OutOfRangeException $e ) { }
						
						return $lang->addToStack( 'transaction_debit' );
				}
				
				return $val;
			},
			'transaction_message' => function( $val, $row )
			{
				return $val;
			},
			'transaction_amount' => function( $val, $row ) use ( $self )
			{
				if ( $val < 0 )
				{
					return "<span style='color:red'>" . $self->currency()->display( $val ) . "</span>";
				}
				else
				{
					return $self->currency()->display( $val );
				}
			},
			'transaction_balance' => function( $val, $row ) use ( $self )
			{
				if ( $val < 0 )
				{
					return "<span style='color:red'>" . $self->currency()->display( $val ) . "</span>";
				}
				else if ( $val > 0 )
				{
					return "<span style='color:green'>" . $self->currency()->display( $val ) . "</span>";
				}
				else
				{
					return $self->currency()->display( $val );
				}
			},
		);
		
		return $table;
	}

	/**
	 * Unclaimed credits table
	 *
	 * @param	int				$limit			Number of credits to show per page
	 * @param	\IPS\Http\Url			$url			The url that the table is being displayed on
	 * @return	\IPS\Helpers\Table\Db
	 */
	public function unclaimedTable( $limit=25, $url=NULL )
	{
		$url = $url ?: \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=unclaimed', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
		
		/* Create the table */
		$table = new \IPS\Helpers\Table\Db( 'points_credits', $url, array( array( 'credit_member_id=? AND credit_currency_id=? AND credit_bank_id=? AND credit_expiration>? AND credit_transaction_id=0', $this->member_id, $this->currency_id, $this->bank_id, time() ) ) );
		
		$table->tableTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'table' );
		$table->rowsTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'rows' );
		
		$table->include = array( 'credit_expiration', 'credit_amount', 'credit_message' );
		$table->langPrefix = 'table_';
		$table->sortBy = 'credit_expiration';
		$table->sortDirection = 'ASC';
		$table->noSort = $table->include;
		$table->limit = $limit;
		$self = $this;
		
		$table->parsers = array
		(
			'credit_timestamp' => function( $val, $row )
			{
				$datetime = \IPS\DateTime::ts( $val );
				return (string) $datetime;
			},
			'credit_message' => function( $val, $row )
			{
				return $val;
			},
			'credit_amount' => function( $val, $row ) use ( $self )
			{
				return $self->currency()->display( $val );
			},
			'credit_expiration' => function( $val, $row ) use ( $self )
			{
				$datetime = \IPS\DateTime::ts( $val );
				return (string) $datetime;
			},
		);
		
		$table->rowButtons = function( $row ) use ( $self )
		{
			$buttons = array();
			$credit = \IPS\points\Transaction\Credit::constructFromData( $row );
			
			if ( \IPS\Member::loggedIn()->member_id === $credit->member()->member_id )
			{
				$buttons[ 'claim' ] = array(
					'icon'		=> 'plus-circle',
					'title'		=> 'points_claim',
					'id'		=> "{$credit->id}-claim",
					'class'		=> "ipsButton ipsButton_veryVerySmall ipsButton_fullWidth ipsButton_positive",
					'link'		=> \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=claim', 'front' )->setQueryString( array( 'credit' => $credit->id ) ),
					'data'		=> array( 'confirm' => TRUE ),
				);
			}
			
			return $buttons;
		};
		
		return $table;
	}
	
	/**
	 * Unpaid debits table
	 *
	 * @param	int				$limit			Number of credits to show per page
	 * @param	\IPS\Http\Url			$url			The url that the table is being displayed on
	 * @return	\IPS\Helpers\Table\Db
	 */
	public function unpaidTable( $limit=25, $url=NULL )
	{
		$url = $url ?: \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=unpaid', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
		
		/* Create the table */
		$table = new \IPS\Helpers\Table\Db( 'points_debits', $url, array( array( 'debit_member_id=? AND debit_currency_id=? AND debit_bank_id=? AND debit_expiration>? AND debit_transaction_id=0', $this->member_id, $this->currency_id, $this->bank_id, time() ) ) );
		
		$table->tableTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'table' );
		$table->rowsTemplate = array( \IPS\Theme::i()->getTemplate( 'tables', 'core', 'admin' ), 'rows' );
		
		$table->include = array( 'debit_expiration', 'debit_amount', 'debit_message' );
		$table->langPrefix = 'table_';
		$table->sortBy = 'debit_expiration';
		$table->sortDirection = 'ASC';
		$table->noSort = $table->include;
		$table->limit = $limit;
		$self = $this;
		
		$table->parsers = array
		(
			'debit_timestamp' => function( $val, $row )
			{
				$datetime = \IPS\DateTime::ts( $val );
				return (string) $datetime;
			},
			'debit_message' => function( $val, $row )
			{
				return $val;
			},
			'debit_amount' => function( $val, $row ) use ( $self )
			{
				return $self->currency()->display( $val );
			},
			'debit_expiration' => function( $val, $row ) use ( $self )
			{
				$datetime = \IPS\DateTime::ts( $val );
				return (string) $datetime;
			},
		);
		
		$table->rowButtons = function( $row ) use ( $self )
		{
			$buttons = array();
			$debit = \IPS\points\Transaction\Debit::constructFromData( $row );
			
			if ( \IPS\Member::loggedIn()->member_id === $debit->member()->member_id )
			{
				$buttons[ 'claim' ] = array(
					'icon'		=> 'credit-card',
					'title'		=> 'points_pay',
					'id'		=> "{$debit->id}-pay",
					'class'		=> "ipsButton ipsButton_veryVerySmall ipsButton_fullWidth ipsButton_normal",
					'link'		=> \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=pay', 'front' )->setQueryString( array( 'debit' => $debit->id ) ),
					'data'		=> array( 'confirm' => TRUE ),
				);
			}
			
			return $buttons;
		};
		
		return $table;
	}
	
	/** 
	 * Get a chart of the balance over a period of time
	 *
	 * @param	string				$history		Period of history to show on chart
	 * @param	string				$plotInterval		Period between each plotted point
	 * @param	string				$dateFormat		Date format to show on chart
	 * @param	\IPS\DateTime			$date			The ending date/time of the chart
	 * @return	\IPS\Helpers\Chart
	 */
	public function balanceChart( $history='P7D', $plotInterval='P1D', $dateFormat='n/j/Y', \IPS\DateTime $date=NULL )
	{
		$chart = new \IPS\points\Chart;
		$member = \IPS\Member::loggedIn();
		
		$chart->addHeader( $member->language()->addToStack( 'points_chart_date' ), 'string' );
		$chart->addHeader( $member->language()->addToStack( 'points_chart_balance' ), 'number' );
		
		$date = $date ?: ( $member->timezone ? new \IPS\DateTime( 'tomorrow -1 second', new \DateTimeZone( $member->timezone ) ) : new \IPS\DateTime( 'tomorrow -1 second' ) );
		$endAt = $date->getTimestamp();
		
		$date->sub( new \DateInterval( $history ) );

		while ( $date->getTimestamp() <= $endAt )
		{	
			$chart->addRow( array( $date->format( $dateFormat ), $this->currency()->round( $this->balanceAsOf( $date ) ) ) );
			$date->add( new \DateInterval( $plotInterval ) );
		}
		
		return $chart;
	}
	
	/**
	 * Get balance as of a certain date/time
	 *
	 * @param	\IPS\DateTime		$date			The date/time to retrieve the balance for
	 * @return	int|float
	 */
	public function balanceAsOf( \IPS\DateTime $date )
	{
		if ( $this->id )
		{
			
			try
			{
				$balance = \IPS\Db::i()->select( 'transaction_balance', 'points_transactions', array( 'transaction_balance_id=? AND transaction_timestamp <=?', $this->id, $date->getTimestamp() ), 'transaction_timestamp DESC', 1 )->first();
				return $this->currency()->smallest( $balance );
			}
			catch( \UnderflowException $e ) { }
		}
		
		return 0;
	}
	
	/**
	 * Can be transfered somewhere?
	 *
	 * @return 	bool
	 */
	public function isTransferable()
	{
		/* Can bank balance be transferred to cash account */
		if ( $this->bank() )
		{
			return $this->bank()->can( 'transfer_out', $this->member() );
		}
		else
		{
			foreach( \IPS\points\Bank::roots( 'transfer_in', $this->member() ) as $bank )
			{
				if ( $bank->canUse( $this->currency(), $this->member() ) )
				{
					return TRUE;
				}
			}
		}
		
		return FALSE;
	}
	
	/**
	 * Can be exchanged?
	 *
	 * @return	bool
	 */
	public function isExchangeable()
	{
		if ( $this->currency()->can( 'exchange_out', $this->member() ) )
		{
			$currencies = $this->bank() ? $this->bank()->currencies( 'exchange_in', $this->member() ) : \IPS\points\Currency::roots( 'exchange_in', $this->member() );
			
			foreach( $currencies as $currency )
			{
				if ( $currency->id !== $this->currency()->id )
				{
					return TRUE;
				}
			}
		}
		
		return FALSE;
	}
	
	/**
	 * Can be sent?
	 *
	 * @return	bool
	 */
	public function isSendable()
	{
		if ( $this->bank() )
		{
			return $this->currency()->can( 'send', $this->member() ) and $this->bank()->can( 'transfer_out', $this->member() );
		}
		else
		{
			return $this->currency()->can( 'send', $this->member() );
		}
	}
	
	/**
	 * Url
	 *
	 * @param	string			$location		Which url to return
	 * @return	\IPS\Http\Url
	 */
	public function url( $location=NULL )
	{
		switch( $location )
		{
			case 'transfer'		: return \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=transfer', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
			case 'exchange'		: return \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=exchange', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
			case 'send'		: return \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=send', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
			case 'transactions'	: return \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=transactions', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
			default			: return \IPS\Http\Url::internal( 'app=points&module=points&controller=dashboard&do=balanceDashboard', 'front' )->setQueryString( array( 'curr' => $this->currency_id, 'bank' => $this->bank_id ) );
		}
	}
	
	/**
	 * Delete
	 *
	 * @return	void
	 */
	public function delete()
	{
		/* Delete transactions */
		\IPS\Db::i()->delete( 'points_transactions', array( 'transaction_balance_id=?', $this->id ) );
		
		parent::delete();
	}
}
