<?php
/**
 * @brief		Nexus Points Payment Gateway
 * @author		Kevin Carwile (http://www.linkedin.com/in/kevincarwile)
 * @copyright		(c) 2015 - Kevin Carwile
 * @package		Points
 * @since		19 Nov 2015
 */

namespace IPS\points\Nexus;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * Points Gateway
 */
class _Gateway extends \IPS\nexus\Gateway
{
	/* !Features */
	
	const SUPPORTS_REFUNDS = TRUE;
	const SUPPORTS_PARTIAL_REFUNDS = TRUE;
	
	/**
	 * Admin can manually charge using this gateway?
	 *
	 * @return	bool
	 */
	public function canAdminCharge()
	{
		return TRUE;
	}

	/**
	 * Check the gateway can process this...
	 *
	 * @param	$amount			\IPS\nexus\Money	The amount
	 * @param	$billingAddress	\IPS\GeoLocation	The billing address
	 * @return	bool
	 */
	public function checkValidity( \IPS\nexus\Money $amount, \IPS\GeoLocation $billingAddress=NULL, \IPS\nexus\Customer $customer=NULL )
	{
		$settings = json_decode( $this->settings, TRUE );
		
		try
		{
			$currency = \IPS\points\Currency::load( $settings[ 'gateway_payment_source' ] );
		}
		catch( \OutOfRangeException $e )
		{
			return FALSE;
		}
		
		return $currency->can( 'use' ) and parent::checkValidity( $amount, $billingAddress, $customer );
	}

	/* !Payment Gateway */
		
	/**
	 * Capture
	 *
	 * @param	\IPS\nexus\Transaction	$transaction	Transaction
	 * @return	void
	 * @throws	\LogicException
	 */
	public function capture( \IPS\nexus\Transaction $transaction )
	{
		$settings = json_decode( $this->settings, TRUE );
		
		try
		{
			$balance = \IPS\points\Balance::get( $transaction->member, $settings[ 'gateway_payment_source' ] );
			$transaction_amount = (string) $transaction->amount->amount;
			
			if ( ! $settings[ 'gateway_rate_' . $transaction->amount->currency ] )
			{
				throw new \OutOfRangeException;
			}
			
			$point_amount = $transaction_amount * $settings[ 'gateway_rate_' . $transaction->amount->currency ];
			
			/* Authorize */
			$balance->canDebit( $point_amount, TRUE );
			
			/* Capture */
			$balance->debit( $point_amount, $transaction->member->language()->get( 'points_nexus_payment' ) . ': ' . $transaction->invoice->title );
			
			return TRUE;
		}
		catch( \IPS\points\Exception $e )
		{
			throw new \OutOfRangeException( $e->getMessage() );
		}
	}
	
	/**
	 * Refund
	 *
	 * @param	\IPS\nexus\Transaction	$transaction	Transaction to be refunded
	 * @param	float|NULL				$amount			Amount to refund (NULL for full amount - always in same currency as transaction)
	 * @return	mixed									Gateway reference ID for refund, if applicable
	 * @throws	\Exception
 	 */
	public function refund( \IPS\nexus\Transaction $transaction, $amount = NULL )
	{
		$settings = json_decode( $this->settings, TRUE );
		
		try
		{
			$balance = \IPS\points\Balance::get( $transaction->member, $settings[ 'gateway_payment_source' ] );
			$transaction_amount = (string) $transaction->amount->amount;
			
			if ( ! $settings[ 'gateway_rate_' . $transaction->amount->currency ] )
			{
				throw new \OutOfRangeException;
			}
			
			$point_amount = $transaction_amount * $settings[ 'gateway_rate_' . $transaction->amount->currency ];
			
			/* Refund */
			$balance->credit( $point_amount, $transaction->member->language()->get( 'points_nexus_refund' ) . ': ' . $transaction->invoice->title, 0, array(), FALSE );			
		}
		catch( \IPS\points\Exception $e )
		{
			throw new \OutOfRangeException( $e->getMessage() );
		}		
	}

	/* !ACP Configuration */
	
	/**
	 * Settings
	 *
	 * @param	\IPS\Helpers\Form	$form	The form
	 * @return	void
	 */
	public function settings( &$form )
	{
		$settings = json_decode( $this->settings, TRUE ) ?: array();

		$points_types = array();
		foreach( \IPS\points\Currency::roots( NULL ) as $currency )
		{
			$points_types[ $currency->id ] = $currency->_title;
		}
		
		$form->add( new \IPS\Helpers\Form\Radio( 'points_gateway_payment_source', $settings[ 'gateway_payment_source' ], TRUE, array( 'options' => $points_types ) ) );
		
		foreach( \IPS\nexus\Money::currencies() as $currency_code )
		{
			\IPS\Member::loggedIn()->language()->words[ 'points_gateway_rate_' . $currency_code ] = $currency_code;
			\IPS\Member::loggedIn()->language()->words[ 'points_gateway_rate_' . $currency_code . '_desc' ] = \IPS\Member::loggedIn()->language()->get( 'points_gateway_rate_desc' );
			$form->add( new \IPS\Helpers\Form\Number( 'points_gateway_rate_' . $currency_code, $settings[ 'gateway_rate_' . $currency_code ] ?: 1, TRUE, array( 'min' => 0.00001, 'decimals' => 5 ) ) );
		}
		
	}
	
	/**
	 * Test Settings
	 *
	 * @param	array	$settings	Settings
	 * @return	array
	 * @throws	\InvalidArgumentException
	 */
	public function testSettings( $settings )
	{
		return $settings;
	}
	
}