<?php
/**
 * @brief		Member Points Systems
 * @author		Kevin Carwile (http://www.linkedin.com/in/kevincarwile)
 * @copyright		(c) 2015 - Kevin Carwile
 * @package		Points
 * @since		19 Nov 2015
 */

namespace IPS\points;

/* To prevent PHP errors (extending class does not exist) revealing path */
if ( !defined( '\IPS\SUITE_UNIQUE_KEY' ) )
{
	header( ( isset( $_SERVER['SERVER_PROTOCOL'] ) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0' ) . ' 403 Forbidden' );
	exit;
}

/**
 * Points Transaction
 */
abstract class _Transaction extends \IPS\Patterns\ActiveRecord
{

	/**
	 * Issue the transaction
	 *
	 * @param	string		$message		Message to record with the transaction
	 * @param	array		$data			Data to record with the transaction
	 * @param	bool		$enforcePerms		Enforce permissions
	 * @return	void
	 *
	 * @throws	\IPS\points\Exception
	 */
	public function issue( $message='', $data=array(), $enforcePerms=TRUE )
	{
		if ( $this->transaction_id == 0 )
		{
			if ( $this instanceof \IPS\points\Transaction\Credit )
			{
				/* Throw exception for expired unclaimed credits */
				if ( $this->expiration > 0 and $this->expiration < time() )
				{
					throw new \IPS\points\Exception( \IPS\Member::loggedIn()->language()->get( 'points_exception_credit_expired' ) );
				}
			}
			
			/* Adjust the balance */
			$balance = \IPS\points\Balance::get( $this->member_id, $this->currency_id, $this->bank_id, $enforcePerms );
			$balance->adjust( $this->amount, $enforcePerms );
			
			/* Log the transaction */
			$transaction = new \IPS\points\Transaction\Register;
			$transaction->balance_id = $balance->id;
			$transaction->balance = $balance->amount;
			$transaction->amount = $this->amount;
			$transaction->message = $message;
			$transaction->data = json_encode( $data );
			$transaction->timestamp = time();
			$transaction->type = $this->type;
			$transaction->parent_id = $this->id;
			$transaction->save();
			
			/* Link the transaction */
			$this->transaction_id = $transaction->id;
			$this->save();
		}
	}
	
	/**
	 * Expire Transaction
	 */
	public function expire()
	{
		/* Expire credits/debits and reverse non-claimed transfers */
		if ( $this->transaction_id == 0 )
		{
			$this->transaction_id = -1;
			$this->save();
			
			if ( $this instanceof \IPS\points\Transaction\Credit and $transfer = $this->transfer() )
			{
				try
				{
					$debit 		= $transfer->debit();
					$return_balance = $debit->balance( FALSE );
					$reverselang 	= $debit->member()->language()->get( 'points_transfer_reversed' );
					
					$return_balance->credit( $debit->amount, $debit->message ? $reverselang . ': ' . $debit->message : $reverselang, 0, array( 'reversed' => $transfer->id ), FALSE );
				}
				catch( \Exception $e ) { }
			}
			
			/* Trigger rules event */
			if ( \IPS\Application::appIsEnabled( 'rules' ) )
			{
				if ( $this instanceof \IPS\points\Transaction\Credit )
				{
					try
					{
						$balance = $this->balance( FALSE );
						\IPS\rules\Event::load( 'points', 'Points', 'credit_expired' )->trigger( $balance, $this->amount );
					}
					catch( \IPS\points\Exception $e ) { }
				}
			}
		}
	}
	
	/**
	 * Get Transaction Register
	 *
	 * @return 	\IPS\points\Transaction\Register|NULL
	 */
	public function register()
	{
		if ( $this->transaction_id )
		{
			return \IPS\points\Transaction\Register::load( $this->transaction_id );
		}
		
		return NULL;
	}
	
	/**
	 * Get Transaction Currency
	 */
	public function currency()
	{
		return \IPS\points\Currency::load( $this->currency_id );
	}
	
	/**
	 * Get Transaction Member
	 */
	public function member()
	{
		return \IPS\Member::load( $this->member_id );
	}
	
	/**
	 * Get Transaction Bank
	 */
	public function bank()
	{
		if ( $this->bank_id )
		{
			return \IPS\points\Bank::load( $this->bank_id );
		}
		
		return NULL;
	}
	
	/**
	 * Get Balance
	 */
	public function balance( $enforcePerms=TRUE )
	{
		return \IPS\points\Balance::get( $this->member_id, $this->currency_id, $this->bank_id, $enforcePerms );
	}
	
	/**
	 * Get Transfer
	 */
	public function transfer()
	{
		if ( $this->transfer_id )
		{
			return \IPS\points\Transaction\Transfer::load( $this->transfer_id );
		}
		
		return NULL;
	}

}
